# -*- coding: utf-8 -*-
#
# Copyright 2012-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""TestReader class

   Manages the test input data, its headers and checks them against the
   model fields data to build the dict input_data.

"""


import sys
import os

from bigml.util import get_csv_delimiter, is_image
from bigml.io import UnicodeReader

from bigmler.utils import BIGML_SYS_ENCODING
from bigmler.checkpoint import file_number_of_lines
from bigmler.utf8recoder import UTF8Recoder
from bigmler.folderreader import FolderReader


def is_folder(folder):
    """Checking if the test_set is a folder """
    if isinstance(folder, str) and os.path.isdir(folder) and \
            os.path.exists(folder):
        return True
    return False


def contains_csv(folder):
    """Checking whether a folder contains a CSV file"""
    files = os.listdir(folder)
    return any(os.path.splitext(filename)[1].replace(".", "").lower() == "csv"
               for filename in files)


class TstReader():
    """Retrieves csv info and builds a input data dict

    """
    def __init__(self, test_set, test_set_header, fields, objective_field,
                 test_separator=None):
        """Builds a generator from a csv file and the fields' model structure

           `test_set`: path to the test data file
           `test_set_header`: boolean, True means that headers are first row
                              in the file
           `fields`: Fields object with the expected fields structure.
           `objective_field`: field_id of the objective field
        """
        self.test_set = test_set
        if isinstance(self.test_set, str) and os.path.isdir(self.test_set) \
                and os.path.exists(self.test_set) and \
                not self.test_set.endswith(os.sep):
            self.test_set += os.sep
        self.directory = None
        if test_set.__class__.__name__ == "StringIO":
            self.test_set = UTF8Recoder(test_set, BIGML_SYS_ENCODING)
        else:
            self.directory = os.path.dirname(self.test_set)
        self.test_set_header = test_set_header
        self.fields = fields
        if (objective_field is not None and
                not objective_field in fields.fields):
            try:
                objective_field = fields.field_id(objective_field)
            except ValueError as exc:
                sys.exit(exc)
        self.objective_field = objective_field
        self.image_fields = []
        try:
            self.image_fields = [column for column in
                                 sorted(fields.fields_by_column_number.keys())
                                 if fields.fields[fields.field_id(column)].get(
                                    "optype") == "image"]
        except ValueError as exc:
            sys.exit(exc)

        self.test_separator = (test_separator
                               if test_separator is not None
                               else get_csv_delimiter())

        if is_folder(self.test_set) and \
                self.image_fields and len(self.image_fields) == 1 \
                and not contains_csv(self.test_set):
            # The test_set points to a directory where images are stored.
            # Only images are read.
            image_field_name = fields.fields[
                fields.field_id(self.image_fields[0])].get(
                "name") if test_set_header else None
            self.test_reader = FolderReader(
                self.test_set,
                filter_fn=is_image,
                header=image_field_name).open_reader()
        else:
            if len(self.test_separator) > 1:
                sys.exit("Only one character can be used as test data "
                         "separator.")
            try:
                self.test_reader = UnicodeReader(
                    self.test_set, delimiter=self.test_separator,
                    lineterminator="\n").open_reader()
            except IOError:
                sys.exit("Error: cannot read test %s" % test_set)

        self.headers = None
        self.raw_headers = None
        self.exclude = []

        if test_set_header:
            self.headers = next(self.test_reader)
            # validate headers against model fields excluding objective_field,
            # that may be present or not
            if objective_field is not None:
                objective_field = fields.field_column_number(objective_field)
            try:
                fields_names = [fields.fields[fields.field_id(i)]
                                ['name'] for i in
                                sorted(fields.fields_by_column_number.keys())
                                if objective_field is None or
                                i != objective_field]
            except ValueError as exc:
                sys.exit(exc)
            self.raw_headers = self.headers[:]

            self.exclude = [i for i in range(len(self.headers))
                            if not self.headers[i] in fields_names]

            self.exclude.reverse()
            if self.exclude:
                if len(self.headers) > len(self.exclude):
                    for index in self.exclude:
                        del self.headers[index]
                else:
                    raise ValueError(("No test field matches the model fields."
                                      "\nThe expected fields are:\n\n%s\n\n"
                                      "while "
                                      "the headers found in the test file are:"
                                      "\n\n%s\n\n"
                                      "Use --no-test-header flag if first li"
                                      "ne should not be interpreted as"
                                      " headers." %
                                      (",".join(fields_names),
                                       ",".join(self.headers))).encode("utf-8"))
        else:
            columns = fields.fields_columns[:]
            if objective_field is not None:
                if fields.field_column_number(objective_field) in columns:
                    columns.remove(fields.field_column_number(objective_field))
            self.headers = [fields.fields_by_column_number[column] for
                            column in columns]
            self.raw_headers = self.headers

    def __iter__(self):
        """Iterator method

        """
        return self

    def __next__(self):
        """Returns the next row

        """
        row = next(self.test_reader)
        if self.directory and self.image_fields:
            for index, row_item in enumerate(row):
                if index in self.image_fields:
                    row[index] = os.path.join(self.directory, row_item)
        return row

    def dict(self, row, filtering=True):
        """Returns the row in a dict format according to the given headers

        """
        new_row = row[:]
        if not filtering:
            if self.test_set_header:
                headers = self.raw_headers
            else:
                headers = [self.fields.fields_by_column_number[column] for
                           column in self.fields.fields_columns]
            return dict(list(zip(headers, new_row)))
        for index in self.exclude:
            del new_row[index]
        return self.fields.pair(new_row, self.headers, self.objective_field)

    def number_of_tests(self):
        """Returns the number of tests in the test file

        """
        tests = file_number_of_lines(self.test_set)
        if self.test_set_header:
            tests -= 1
        return tests

    def has_headers(self):
        """Returns whether the test set file has a headers row

        """
        return self.test_set_header

    def close(self):
        """Closing file handler

        """
        try:
            self.test_reader.close_reader()
        except AttributeError:
            pass



# -*- coding: utf-8 -*-
#
# Copyright 2020-2025 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Resources management functions

"""


import sys

import bigml.api

from bigmler.utils import (dated, get_url, log_message, check_resource,
                           plural, is_shared,
                           check_resource_error, log_created_resources)
from bigmler.reports import report

from bigmler.resourcesapi.common import set_basic_model_args, \
    configure_input_fields, update_sample_parameters_args, \
    update_json_args, wait_for_available_tasks, get_basic_seed, \
    relative_input_fields, get_all_labels, label_model_args

from bigmler.resourcesapi.common import SEED, FIELDS_QS, \
    ALL_FIELDS_QS, EVALUATE_SAMPLE_RATE


def set_model_args(args, name=None, objective_id=None, fields=None,
                   model_fields=None, other_label=None):
    """Return model arguments dict

    """
    if name is None:
        name = args.name
    if objective_id is None and args.max_categories is None:
        objective_id = args.objective_id_
    if args.max_categories > 0:
        objective_id = args.objective_field
    if model_fields is None:
        model_fields = args.model_fields_

    model_args = set_basic_model_args(args, name)
    model_args.update({"missing_splits": args.missing_splits})
    if objective_id is not None and fields is not None:
        model_args.update({"objective_field": objective_id})

    # If evaluate flag is on and no test_split flag is provided,
    # we choose a deterministic sampling with
    # args.sample_rate (80% by default) of the data to create the model
    # If cross_validation_rate = n/100, then we choose to run 2 * n evaluations
    # by holding out a n% of randomly sampled data.

    if ((args.evaluate and args.test_split == 0 and args.test_datasets is None)
            or args.cross_validation_rate > 0):
        model_args.update(seed=SEED)
        if args.cross_validation_rate > 0:
            args.sample_rate = 1 - args.cross_validation_rate
            args.replacement = False
        elif (args.sample_rate == 1 and args.test_datasets is None
              and not args.dataset_off):
            args.sample_rate = EVALUATE_SAMPLE_RATE
    if model_fields and fields is not None:
        use_name = args.max_categories > 0
        input_fields = configure_input_fields(
            fields, model_fields, by_name=use_name)
        model_args.update(input_fields=input_fields)

    if args.pruning and args.pruning != 'smart':
        stat_pruning = args.pruning == 'statistical'
        model_args.update(stat_pruning=stat_pruning)

    if args.node_threshold > 0:
        model_args.update(node_threshold=args.node_threshold)

    if args.balance:
        model_args.update(balance_objective=True)

    if args.split_field:
        model_args.update(split_field=args.split_field)

    if args.focus_field:
        model_args.update(focus_field=args.focus_field)

    if args.weight_field:
        try:
            weight_field = fields.field_id(args.weight_field)
        except ValueError as exc:
            sys.exit(exc)
        model_args.update(weight_field=weight_field)

    if args.objective_weights:
        model_args.update(objective_weights=args.objective_weights_json)

    if args.max_categories > 0:
        model_args.update(
            user_metadata={'other_label': other_label,
                           'max_categories': args.max_categories})

    model_args = update_sample_parameters_args(model_args, args)

    if 'model' in args.json_args:
        update_json_args(model_args, args.json_args.get('model'), fields)

    return model_args


def set_label_model_args(args, fields, labels, multi_label_data):
    """Set of args needed to build a model per label

    """

    objective_field = args.objective_field
    if not args.model_fields_:
        model_fields = []
    else:
        model_fields = relative_input_fields(fields, args.model_fields_)
    if objective_field is None:
        objective_field = fields.objective_field
    try:
        objective_id = fields.field_id(objective_field)
        objective_field = fields.field_name(objective_id)
    except ValueError as exc:
        sys.exit(exc)
    all_labels = get_all_labels(multi_label_data)
    model_args_list = []

    for index in range(args.number_of_models - 1, -1, -1):
        label = labels[index]
        (new_name, label_field, single_label_fields) = label_model_args(
            args.name, label, all_labels, model_fields,
            objective_field)
        model_args = set_model_args(args, name=new_name,
                                    objective_id=label_field, fields=fields,
                                    model_fields=single_label_fields)
        if multi_label_data is not None:
            model_args.update(
                user_metadata={'multi_label_data': multi_label_data})
        model_args_list.append(model_args)
    return model_args_list


def create_models(datasets, model_ids, model_args,
                  args, api=None, path=None,
                  session_file=None, log=None):
    """Create remote models

    """
    if api is None:
        api = bigml.api.BigML()

    models = model_ids[:]
    existing_models = len(models)
    model_args_list = []
    if args.dataset_off and args.evaluate:
        args.test_dataset_ids = datasets[:]
    if not args.multi_label:
        datasets = datasets[existing_models:]
    # if resuming and all models were created, there will be no datasets left
    if datasets:
        dataset = datasets[0]
        if isinstance(model_args, list):
            model_args_list = model_args
        if args.number_of_models > 0:
            message = dated("Creating %s.\n" %
                            plural("model", args.number_of_models))
            log_message(message, log_file=session_file,
                        console=args.verbosity)

            single_model = args.number_of_models == 1 and existing_models == 0
            # if there's more than one model the first one must contain
            # the entire field structure to be used as reference.
            query_string = (FIELDS_QS if single_model and (args.test_header \
                and not args.export_fields) else ALL_FIELDS_QS)
            inprogress = []
            for i in range(0, args.number_of_models):
                wait_for_available_tasks(inprogress, args.max_parallel_models,
                                         api, "model")
                if model_args_list:
                    model_args = model_args_list[i]
                if args.cross_validation_rate > 0:
                    new_seed = get_basic_seed(i + existing_models)
                    model_args.update(seed=new_seed)
                # one model per dataset (--max-categories or single model)
                if (args.max_categories > 0 or
                        (args.test_datasets and args.evaluate)):
                    dataset = datasets[i]
                    model = api.create_model(dataset, model_args, retries=None)
                elif args.dataset_off and args.evaluate:
                    multi_dataset = args.test_dataset_ids[:]
                    del multi_dataset[i + existing_models]
                    model = api.create_model(multi_dataset, model_args,
                                             retries=None)
                else:
                    model = api.create_model(datasets, model_args,
                                             retries=None)
                model_id = check_resource_error(model,
                                                "Failed to create model: ")
                log_message("%s\n" % model_id, log_file=log)
                model_ids.append(model_id)
                inprogress.append(model_id)
                models.append(model)
                log_created_resources("models", path, model_id, mode='a')

            if args.number_of_models < 2 and args.verbosity:
                if bigml.api.get_status(model)['code'] != bigml.api.FINISHED:
                    try:
                        model = check_resource(model, api.get_model,
                                               query_string=query_string,
                                               raise_on_error=True)
                    except Exception as exception:
                        sys.exit("Failed to get a finished model: %s" %
                                 str(exception))
                    models[0] = model
                message = dated("Model created: %s\n" %
                                get_url(model))
                log_message(message, log_file=session_file,
                            console=args.verbosity)
                if args.reports:
                    report(args.reports, path, model)

    return models, model_ids


def create_model(cluster, model_args, args, api=None,
                 path=None, session_file=None, log=None, model_type=None):
    """Creates remote model from cluster and centroid

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Creating model.\n")
    log_message(message, log_file=session_file, console=args.verbosity)
    model = api.create_model(cluster, model_args, retries=None)
    suffix = "" if model_type is None else "_%s" % model_type
    log_created_resources("models%s" % suffix, path,
                          bigml.api.get_model_id(model), mode='a')
    model_id = check_resource_error(model, "Failed to create model: ")
    try:
        model = check_resource(model, api.get_model,
                               query_string=ALL_FIELDS_QS,
                               raise_on_error=True)
    except Exception as exception:
        sys.exit("Failed to get a finished model: %s" % str(exception))
    message = dated("Model created: %s\n" % get_url(model))
    log_message(message, log_file=session_file, console=args.verbosity)
    log_message("%s\n" % model_id, log_file=log)
    if args.reports:
        report(args.reports, path, model)
    return model


def update_model(model, model_args, args,
                 api=None, path=None, session_file=None):
    """Updates model properties

    """
    if api is None:
        api = bigml.api.BigML()
    message = dated("Updating model. %s\n" %
                    get_url(model))
    log_message(message, log_file=session_file,
                console=args.verbosity)
    model = api.update_model(model, model_args)
    check_resource_error(model, "Failed to update model: %s"
                         % model['resource'])
    model = check_resource(model, api.get_model, query_string=ALL_FIELDS_QS,
                           raise_on_error=True)
    if is_shared(model):
        message = dated("Shared model link. %s\n" %
                        get_url(model, shared=True))
        log_message(message, log_file=session_file, console=args.verbosity)
        if args.reports:
            report(args.reports, path, model)

    return model


def get_models(model_ids, args, api=None, session_file=None):
    """Retrieves remote models in its actual status

    """
    if api is None:
        api = bigml.api.BigML()
    model_id = ""
    models = model_ids
    single_model = len(model_ids) == 1
    if single_model:
        model_id = model_ids[0]
    message = dated("Retrieving %s. %s\n" %
                    (plural("model", len(model_ids)),
                     get_url(model_id)))
    log_message(message, log_file=session_file, console=args.verbosity)
    if len(model_ids) < args.max_batch_models:
        models = []
        for model in model_ids:
            try:
                # if there's more than one model the first one must contain
                # the entire field structure to be used as reference.
                query_string = (
                    ALL_FIELDS_QS if (
                        (not single_model and (not models or
                                               args.multi_label)) or
                        not args.test_header)
                    else FIELDS_QS)
                model = check_resource(model, api.get_model,
                                       query_string=query_string,
                                       raise_on_error=True)
            except Exception as exception:
                sys.exit("Failed to get a finished model: %s" %
                         str(exception))
            models.append(model)
        model = models[0]
    else:
        try:
            query_string = (ALL_FIELDS_QS if not single_model or
                            not args.test_header
                            else FIELDS_QS)
            model = check_resource(model_ids[0], api.get_model,
                                   query_string=query_string,
                                   raise_on_error=True)
        except Exception as exception:
            sys.exit("Failed to get a finished model: %s" % str(exception))
        models[0] = model

    return models, model_ids


def set_publish_model_args(args):
    """Set args to publish model

    """
    public_model = {}
    if args.black_box:
        public_model = {"private": False}
    if args.white_box:
        public_model = {"private": False, "white_box": True}
        if args.model_price:
            public_model.update(price=args.model_price)
        if args.cpp:
            public_model.update(credits_per_prediction=args.cpp)
    return public_model



# -*- coding: utf-8 -*-
def predict_rating(gender=None,
                   age_range=None,
                   occupation=None,
                   zipcode=None,
                   title=None,
                   genres=None,
                   timestamp=None):
    """ Predictor for rating

        Created using BigMLer
    """

    import re

    tm_tokens = 'tokens_only'
    tm_full_term = 'full_terms_only'
    tm_all = 'all'

    def term_matches(text, field_name, term):
        """ Counts the number of occurences of term and its variants in text

        """
        forms_list = term_forms[field_name].get(term, [term])
        options = term_analysis[field_name]
        token_mode = options.get('token_mode', tm_tokens)
        case_sensitive = options.get('case_sensitive', False)
        first_term = forms_list[0]
        if token_mode == tm_full_term:
            return full_term_match(text, first_term, case_sensitive)
        else:
            # In token_mode='all' we will match full terms using equals and
            # tokens using contains
            if token_mode == tm_all and len(forms_list) == 1:
                pattern = re.compile(r'^.+\b.+$', re.U)
                if re.match(pattern, first_term):
                    return full_term_match(text, first_term, case_sensitive)
            return term_matches_tokens(text, forms_list, case_sensitive)


    def full_term_match(text, full_term, case_sensitive):
        """Counts the match for full terms according to the case_sensitive
              option

        """
        if not case_sensitive:
            text = text.lower()
            full_term = full_term.lower()
        return 1 if text == full_term else 0

    def get_tokens_flags(case_sensitive):
        """Returns flags for regular expression matching depending on text
              analysis options

        """
        flags = re.U
        if not case_sensitive:
            flags = (re.I | flags)
        return flags


    def term_matches_tokens(text, forms_list, case_sensitive):
        """ Counts the number of occurrences of the words in forms_list in
               the text

        """
        flags = get_tokens_flags(case_sensitive)
        expression = r'(\b|_)%s(\b|_)' % '(\\b|_)|(\\b|_)'.join(forms_list)
        pattern = re.compile(expression, flags=flags)
        matches = re.findall(pattern, text)
        return len(matches)

    term_analysis = {
        "title": {
            "case_sensitive": False,
            "token_mode": 'all',
        },
    }
    term_forms = {
        "title": {
            "beauty": ['beauty', 'beautiful'],
            "day": ['day', 'days'],
        },
    }

    def item_matches(text, field_name, item):
        """ Counts the number of occurrences of item in text

        """
        options = item_analysis[field_name]
        separator = options.get('separator', ' ')
        regexp = options.get('separator_regexp')
        if regexp is None:
            regexp = r"%s" % re.escape(separator)
        return count_items_matches(text, item, regexp)


    def count_items_matches(text, item, regexp):
        """ Counts the number of occurrences of the item in the text

        """
        expression = r'(^|%s)%s($|%s)' % (regexp, item, regexp)
        pattern = re.compile(expression, flags=re.U)
        matches = re.findall(pattern, text)
        return len(matches)

    item_analysis = {
        "genres": {
            "separator": '$',
        },
    }

    if (genres is None):
        return {"prediction": 3.7, "error": 1.28278}
    if (item_matches(genres, "genres", u"Comedy") > 0):
        if (title is None):
            return {"prediction": 3.39535, "error": 1.57231}
        if (term_matches(title, "title", u"life") > 0):
            return {"prediction":5, "error":0.90777}
        if (term_matches(title, "title", u"life") <= 0):
            if (term_matches(title, "title", u"forrest gump (1994)") > 0):
                return {"prediction":5, "error":1.09624}
            if (term_matches(title, "title", u"forrest gump (1994)") <= 0):
                if (term_matches(title, "title", u"1983") > 0):
                    return {"prediction":5, "error":1.08292}
                if (term_matches(title, "title", u"1983") <= 0):
                    if (zipcode is None):
                        return {"prediction": 3.25316, "error": 1.5086}
                    if (zipcode > 7753):
                        if (item_matches(genres, "genres", u"Horror") > 0):
                            if (timestamp is None):
                                return {"prediction": 2, "error": 5.08228}
                            if (timestamp > 978258115):
                                return {"prediction":1.5, "error":5.26764}
                            if (timestamp <= 978258115):
                                return {"prediction":3, "error":5.08228}
                        if (item_matches(genres, "genres", u"Horror") <= 0):
                            if (timestamp is None):
                                return {"prediction": 3.3913, "error": 1.43342}
                            if (timestamp > 978218758):
                                if (item_matches(genres, "genres", u"Thriller") > 0):
                                    return {"prediction":2, "error":10.53528}
                                if (item_matches(genres, "genres", u"Thriller") <= 0):
                                    if (item_matches(genres, "genres", u"Crime") > 0):
                                        return {"prediction":5, "error":0.9578}
                                    if (item_matches(genres, "genres", u"Crime") <= 0):
                                        if (term_matches(title, "title", u"breakfast club, the (1985)") > 0):
                                            return {"prediction":2, "error":1.31722}
                                        if (term_matches(title, "title", u"breakfast club, the (1985)") <= 0):
                                            if (term_matches(title, "title", u"monty") > 0):
                                                return {"prediction":2, "error":1.28344}
                                            if (term_matches(title, "title", u"monty") <= 0):
                                                if (term_matches(title, "title", u"stand by me (1986)") > 0):
                                                    return {"prediction":5, "error":1.24322}
                                                if (term_matches(title, "title", u"stand by me (1986)") <= 0):
                                                    if (timestamp > 978228710):
                                                        if (item_matches(genres, "genres", u"Musical") > 0):
                                                            return {"prediction":4.5, "error":5.26764}
                                                        if (item_matches(genres, "genres", u"Musical") <= 0):
                                                            if (item_matches(genres, "genres", u"Romance") > 0):
                                                                if (term_matches(title, "title", u"day") > 0):
                                                                    return {"prediction":2, "error":1.38964}
                                                                if (term_matches(title, "title", u"day") <= 0):
                                                                    if (timestamp > 978428301):
                                                                        return {"prediction":4, "error":1.13085}
                                                                    if (timestamp <= 978428301):
                                                                        if (term_matches(title, "title", u"shakespeare in love (1998)") > 0):
                                                                            return {"prediction":4, "error":0.958}
                                                                        if (term_matches(title, "title", u"shakespeare in love (1998)") <= 0):
                                                                            return {"prediction":3, "error":0.36209}
                                                            if (item_matches(genres, "genres", u"Romance") <= 0):
                                                                if (occupation is None):
                                                                    return {"prediction": 3.65385, "error": 1.31541}
                                                                if (occupation == "writer"):
                                                                    return {"prediction":5, "error":1.31541}
                                                                if (occupation != "writer"):
                                                                    if (item_matches(genres, "genres", u"Drama") > 0):
                                                                        if (term_matches(title, "title", u"1997") > 0):
                                                                            return {"prediction":5, "error":1.56826}
                                                                        if (term_matches(title, "title", u"1997") <= 0):
                                                                            return {"prediction":4, "error":0.78413}
                                                                    if (item_matches(genres, "genres", u"Drama") <= 0):
                                                                        if (timestamp > 978298248):
                                                                            if (timestamp > 978298391):
                                                                                if (gender is None):
                                                                                    return {"prediction": 3.6, "error": 1.92072}
                                                                                if (gender == "Female"):
                                                                                    return {"prediction":4, "error":1.35815}
                                                                                if (gender == "Male"):
                                                                                    if (term_matches(title, "title", u"1996") > 0):
                                                                                        return {"prediction":4, "error":2.93426}
                                                                                    if (term_matches(title, "title", u"1996") <= 0):
                                                                                        return {"prediction":3, "error":2.07483}
                                                                            if (timestamp <= 978298391):
                                                                                return {"prediction":5, "error":2.36951}
                                                                        if (timestamp <= 978298248):
                                                                            if (term_matches(title, "title", u"1980") > 0):
                                                                                return {"prediction":2, "error":1.31017}
                                                                            if (term_matches(title, "title", u"1980") <= 0):
                                                                                if (timestamp > 978297750):
                                                                                    if (term_matches(title, "title", u"1999") > 0):
                                                                                        return {"prediction":3, "error":1.14938}
                                                                                    if (term_matches(title, "title", u"1999") <= 0):
                                                                                        return {"prediction":4, "error":0.81274}
                                                                                if (timestamp <= 978297750):
                                                                                    if (term_matches(title, "title", u"1994") > 0):
                                                                                        return {"prediction":4, "error":1.09476}
                                                                                    if (term_matches(title, "title", u"1994") <= 0):
                                                                                        return {"prediction":3, "error":0.44694}
                                                    if (timestamp <= 978228710):
                                                        if (timestamp > 978226820):
                                                            return {"prediction":5, "error":2.93426}
                                                        if (timestamp <= 978226820):
                                                            return {"prediction":4, "error":2.07483}
                            if (timestamp <= 978218758):
                                if (term_matches(title, "title", u"1994") > 0):
                                    return {"prediction":1, "error":1.96692}
                                if (term_matches(title, "title", u"1994") <= 0):
                                    if (timestamp > 978174603):
                                        if (term_matches(title, "title", u"1999") > 0):
                                            if (timestamp > 978200667):
                                                return {"prediction":1, "error":3.89486}
                                            if (timestamp <= 978200667):
                                                if (timestamp > 978196617):
                                                    return {"prediction":3, "error":2.07483}
                                                if (timestamp <= 978196617):
                                                    return {"prediction":2, "error":2.93426}
                                        if (term_matches(title, "title", u"1999") <= 0):
                                            if (occupation is None):
                                                return {"prediction": 3.09091, "error": 1.95519}
                                            if (occupation == "executive/managerial"):
                                                return {"prediction":4, "error":1.38253}
                                            if (occupation != "executive/managerial"):
                                                if (timestamp > 978200651):
                                                    if (term_matches(title, "title", u"bride") > 0):
                                                        return {"prediction":2, "error":2.36951}
                                                    if (term_matches(title, "title", u"bride") <= 0):
                                                        if (timestamp > 978202404):
                                                            return {"prediction":3, "error":1.35815}
                                                        if (timestamp <= 978202404):
                                                            if (term_matches(title, "title", u"batman") > 0):
                                                                return {"prediction":3, "error":2.93426}
                                                            if (term_matches(title, "title", u"batman") <= 0):
                                                                return {"prediction":4, "error":2.07483}
                                                if (timestamp <= 978200651):
                                                    if (item_matches(genres, "genres", u"Romance") > 0):
                                                        return {"prediction":3, "error":2.93426}
                                                    if (item_matches(genres, "genres", u"Romance") <= 0):
                                                        return {"prediction":2, "error":2.07483}
                                    if (timestamp <= 978174603):
                                        if (term_matches(title, "title", u"1985") > 0):
                                            return {"prediction":5, "error":2.93395}
                                        if (term_matches(title, "title", u"1985") <= 0):
                                            if (occupation is None):
                                                return {"prediction": 3.5, "error": 2.34869}
                                            if (occupation == "sales/marketing"):
                                                return {"prediction":4, "error":2.34869}
                                            if (occupation != "sales/marketing"):
                                                if (timestamp > 978174551):
                                                    return {"prediction":4, "error":2.93426}
                                                if (timestamp <= 978174551):
                                                    return {"prediction":3, "error":2.07483}
                    if (zipcode <= 7753):
                        if (item_matches(genres, "genres", u"Drama") > 0):
                            return {"prediction":4, "error":2.60606}
                        if (item_matches(genres, "genres", u"Drama") <= 0):
                            if (timestamp is None):
                                return {"prediction": 1.8, "error": 2.93395}
                            if (timestamp > 978904214):
                                if (term_matches(title, "title", u"1997") > 0):
                                    return {"prediction":3, "error":2.93426}
                                if (term_matches(title, "title", u"1997") <= 0):
                                    return {"prediction":2, "error":2.07483}
                            if (timestamp <= 978904214):
                                return {"prediction":1, "error":2.07461}
    if (item_matches(genres, "genres", u"Comedy") <= 0):
        if (title is None):
            return {"prediction": 3.82843, "error": 1.25974}
        if (term_matches(title, "title", u"1995") > 0):
            if (occupation is None):
                return {"prediction": 2.66667, "error": 3.25095}
            if (occupation == "clerical/admin"):
                return {"prediction":1, "error":3.25095}
            if (occupation != "clerical/admin"):
                if (item_matches(genres, "genres", u"Romance") > 0):
                    return {"prediction":4, "error":2.47964}
                if (item_matches(genres, "genres", u"Romance") <= 0):
                    if (occupation == "writer"):
                        return {"prediction":2, "error":2.03402}
                    if (occupation != "writer"):
                        return {"prediction":3, "error":1.17434}
        if (term_matches(title, "title", u"1995") <= 0):
            if (item_matches(genres, "genres", u"Horror") > 0):
                if (timestamp is None):
                    return {"prediction": 3.35, "error": 2.2498}
                if (timestamp > 978200824):
                    if (timestamp > 978876267):
                        return {"prediction":2, "error":1.97983}
                    if (timestamp <= 978876267):
                        if (item_matches(genres, "genres", u"Thriller") > 0):
                            if (term_matches(title, "title", u"alien") > 0):
                                return {"prediction":4, "error":2.93426}
                            if (term_matches(title, "title", u"alien") <= 0):
                                return {"prediction":3, "error":2.07483}
                        if (item_matches(genres, "genres", u"Thriller") <= 0):
                            if (timestamp > 978268588):
                                if (term_matches(title, "title", u"king") > 0):
                                    return {"prediction":4, "error":2.03402}
                                if (term_matches(title, "title", u"king") <= 0):
                                    return {"prediction":5, "error":1.17434}
                            if (timestamp <= 978268588):
                                if (term_matches(title, "title", u"alien") > 0):
                                    return {"prediction":3, "error":1.56826}
                                if (term_matches(title, "title", u"alien") <= 0):
                                    return {"prediction":4, "error":0.78413}
                if (timestamp <= 978200824):
                    if (occupation is None):
                        return {"prediction": 2.42857, "error": 3.28429}
                    if (occupation == "academic/educator"):
                        if (term_matches(title, "title", u"1960") > 0):
                            return {"prediction":4, "error":2.93426}
                        if (term_matches(title, "title", u"1960") <= 0):
                            return {"prediction":3, "error":2.07483}
                    if (occupation != "academic/educator"):
                        if (term_matches(title, "title", u"bringing") > 0):
                            return {"prediction":3, "error":3.89486}
                        if (term_matches(title, "title", u"bringing") <= 0):
                            if (timestamp > 978200492):
                                return {"prediction":2, "error":2.93426}
                            if (timestamp <= 978200492):
                                return {"prediction":1, "error":2.07483}
            if (item_matches(genres, "genres", u"Horror") <= 0):
                if (gender is None):
                    return {"prediction": 3.92135, "error": 1.21004}
                if (gender == "Male"):
                    if (term_matches(title, "title", u"dick tracy (1990)") > 0):
                        return {"prediction":1, "error":1.29316}
                    if (term_matches(title, "title", u"dick tracy (1990)") <= 0):
                        if (occupation is None):
                            return {"prediction": 3.84892, "error": 1.26101}
                        if (occupation == "writer"):
                            if (timestamp is None):
                                return {"prediction": 3.2, "error": 2.52836}
                            if (timestamp > 978243869):
                                if (item_matches(genres, "genres", u"Romance") > 0):
                                    return {"prediction":4, "error":2.5701}
                                if (item_matches(genres, "genres", u"Romance") <= 0):
                                    if (timestamp > 978246320):
                                        if (timestamp > 978246556):
                                            return {"prediction":2, "error":2.93426}
                                        if (timestamp <= 978246556):
                                            return {"prediction":3, "error":2.07483}
                                    if (timestamp <= 978246320):
                                        return {"prediction":2, "error":1.35815}
                            if (timestamp <= 978243869):
                                if (term_matches(title, "title", u"1994") > 0):
                                    return {"prediction":3, "error":3.32155}
                                if (term_matches(title, "title", u"1994") <= 0):
                                    if (item_matches(genres, "genres", u"Film-Noir") > 0):
                                        return {"prediction":4, "error":2.93426}
                                    if (item_matches(genres, "genres", u"Film-Noir") <= 0):
                                        return {"prediction":4.5, "error":5.26764}
                        if (occupation != "writer"):
                            if (term_matches(title, "title", u"2000") > 0):
                                if (term_matches(title, "title", u"mission") > 0):
                                    return {"prediction":2.5, "error":5.26764}
                                if (term_matches(title, "title", u"mission") <= 0):
                                    if (term_matches(title, "title", u"cell") > 0):
                                        return {"prediction":3, "error":1.09476}
                                    if (term_matches(title, "title", u"cell") <= 0):
                                        if (timestamp is None):
                                            return {"prediction": 3.6, "error": 1.92072}
                                        if (timestamp > 978217782):
                                            return {"prediction":3, "error":1.35815}
                                        if (timestamp <= 978217782):
                                            return {"prediction":4, "error":1.10893}
                            if (term_matches(title, "title", u"2000") <= 0):
                                if (timestamp is None):
                                    return {"prediction": 3.95, "error": 1.26219}
                                if (timestamp > 978298955):
                                    if (timestamp > 1009669148):
                                        if (term_matches(title, "title", u"1997") > 0):
                                            return {"prediction":5, "error":2.93426}
                                        if (term_matches(title, "title", u"1997") <= 0):
                                            return {"prediction":4, "error":2.07483}
                                    if (timestamp <= 1009669148):
                                        if (term_matches(title, "title", u"1989") > 0):
                                            return {"prediction":5, "error":1.59717}
                                        if (term_matches(title, "title", u"1989") <= 0):
                                            if (term_matches(title, "title", u"1990") > 0):
                                                return {"prediction":2, "error":1.16977}
                                            if (term_matches(title, "title", u"1990") <= 0):
                                                if (item_matches(genres, "genres", u"Film-Noir") > 0):
                                                    return {"prediction":4, "error":0.95152}
                                                if (item_matches(genres, "genres", u"Film-Noir") <= 0):
                                                    if (term_matches(title, "title", u"1980") > 0):
                                                        return {"prediction":4, "error":0.77415}
                                                    if (term_matches(title, "title", u"1980") <= 0):
                                                        return {"prediction":3, "error":0.25805}
                                if (timestamp <= 978298955):
                                    if (term_matches(title, "title", u"1987") > 0):
                                        return {"prediction":1, "error":1.28682}
                                    if (term_matches(title, "title", u"1987") <= 0):
                                        if (term_matches(title, "title", u"fight") > 0):
                                            return {"prediction":2, "error":1.23008}
                                        if (term_matches(title, "title", u"fight") <= 0):
                                            if (term_matches(title, "title", u"1993") > 0):
                                                if (timestamp > 978234034):
                                                    return {"prediction":4, "error":0.89387}
                                                if (timestamp <= 978234034):
                                                    return {"prediction":3, "error":0.77411}
                                            if (term_matches(title, "title", u"1993") <= 0):
                                                if (term_matches(title, "title", u"1996") > 0):
                                                    if (occupation == "other"):
                                                        return {"prediction":1, "error":2.43201}
                                                    if (occupation != "other"):
                                                        if (item_matches(genres, "genres", u"Drama") > 0):
                                                            return {"prediction":5, "error":1.38965}
                                                        if (item_matches(genres, "genres", u"Drama") <= 0):
                                                            if (zipcode is None):
                                                                return {"prediction": 3.75, "error": 1.96736}
                                                            if (zipcode > 94327):
                                                                return {"prediction":5, "error":1.96736}
                                                            if (zipcode <= 94327):
                                                                if (item_matches(genres, "genres", u"Thriller") > 0):
                                                                    return {"prediction":5, "error":1.90304}
                                                                if (item_matches(genres, "genres", u"Thriller") <= 0):
                                                                    if (occupation == "executive/managerial"):
                                                                        return {"prediction":3, "error":10.53528}
                                                                    if (occupation != "executive/managerial"):
                                                                        if (zipcode > 58365):
                                                                            return {"prediction":3, "error":0.99163}
                                                                        if (zipcode <= 58365):
                                                                            if (timestamp > 978297836):
                                                                                return {"prediction":3, "error":1.28505}
                                                                            if (timestamp <= 978297836):
                                                                                return {"prediction":4, "error":0.57469}
                                                if (term_matches(title, "title", u"1996") <= 0):
                                                    if (term_matches(title, "title", u"negotiator") > 0):
                                                        return {"prediction":3, "error":0.82118}
                                                    if (term_matches(title, "title", u"negotiator") <= 0):
                                                        if (item_matches(genres, "genres", u"War") > 0):
                                                            if (timestamp > 978201771):
                                                                if (timestamp > 978294214):
                                                                    if (timestamp > 978295884):
                                                                        return {"prediction":4, "error":2.07483}
                                                                    if (timestamp <= 978295884):
                                                                        return {"prediction":5, "error":2.93426}
                                                                if (timestamp <= 978294214):
                                                                    if (timestamp > 978211160):
                                                                        if (timestamp > 978294061):
                                                                            return {"prediction":2, "error":2.03402}
                                                                        if (timestamp <= 978294061):
                                                                            return {"prediction":3, "error":1.17434}
                                                                    if (timestamp <= 978211160):
                                                                        return {"prediction":4, "error":2.47964}
                                                            if (timestamp <= 978201771):
                                                                return {"prediction":5, "error":2.56453}
                                                        if (item_matches(genres, "genres", u"War") <= 0):
                                                            if (occupation == "K-12 student"):
                                                                if (timestamp > 978146981):
                                                                    return {"prediction":4, "error":2.93426}
                                                                if (timestamp <= 978146981):
                                                                    return {"prediction":3, "error":2.07483}
                                                            if (occupation != "K-12 student"):
                                                                if (timestamp > 978201899):
                                                                    if (timestamp > 978215603):
                                                                        if (item_matches(genres, "genres", u"Adventure") > 0):
                                                                            if (zipcode is None):
                                                                                return {"prediction": 4.72727, "error": 1.09872}
                                                                            if (zipcode > 22103):
                                                                                if (term_matches(title, "title", u"1994") > 0):
                                                                                    return {"prediction":4, "error":1.72408}
                                                                                if (term_matches(title, "title", u"1994") <= 0):
                                                                                    if (term_matches(title, "title", u"king") > 0):
                                                                                        return {"prediction":4, "error":1.92072}
                                                                                    if (term_matches(title, "title", u"king") <= 0):
                                                                                        if (term_matches(title, "title", u"jones") > 0):
                                                                                            return {"prediction":4, "error":2.03402}
                                                                                        if (term_matches(title, "title", u"jones") <= 0):
                                                                                            return {"prediction":5, "error":1.17434}
                                                                            if (zipcode <= 22103):
                                                                                return {"prediction":5, "error":0.49136}
                                                                        if (item_matches(genres, "genres", u"Adventure") <= 0):
                                                                            if (timestamp > 978294097):
                                                                                if (term_matches(title, "title", u"1960") > 0):
                                                                                    return {"prediction":3, "error":1.25106}
                                                                                if (term_matches(title, "title", u"1960") <= 0):
                                                                                    if (timestamp > 978294245):
                                                                                        if (timestamp > 978298584):
                                                                                            return {"prediction":5, "error":0.80826}
                                                                                        if (timestamp <= 978298584):
                                                                                            if (term_matches(title, "title", u"terminator") > 0):
                                                                                                return {"prediction":5, "error":1.18675}
                                                                                            if (term_matches(title, "title", u"terminator") <= 0):
                                                                                                if (term_matches(title, "title", u"1994") > 0):
                                                                                                    return {"prediction":5, "error":1.18253}
                                                                                                if (term_matches(title, "title", u"1994") <= 0):
                                                                                                    if (occupation == "scientist"):
                                                                                                        return {"prediction":5, "error":1.13085}
                                                                                                    if (occupation != "scientist"):
                                                                                                        if (term_matches(title, "title", u"1976") > 0):
                                                                                                            return {"prediction":5, "error":0.958}
                                                                                                        if (term_matches(title, "title", u"1976") <= 0):
                                                                                                            return {"prediction":4, "error":0.36209}
                                                                                    if (timestamp <= 978294245):
                                                                                        return {"prediction":5, "error":0.60498}
                                                                            if (timestamp <= 978294097):
                                                                                if (timestamp > 978230086):
                                                                                    if (timestamp > 978234842):
                                                                                        return {"prediction":4, "error":1.09476}
                                                                                    if (timestamp <= 978234842):
                                                                                        if (term_matches(title, "title", u"truman") > 0):
                                                                                            return {"prediction":4, "error":1.56826}
                                                                                        if (term_matches(title, "title", u"truman") <= 0):
                                                                                            return {"prediction":3, "error":0.78413}
                                                                                if (timestamp <= 978230086):
                                                                                    if (term_matches(title, "title", u"graduate, the (1967)") > 0):
                                                                                        return {"prediction":3, "error":1.65457}
                                                                                    if (term_matches(title, "title", u"graduate, the (1967)") <= 0):
                                                                                        if (term_matches(title, "title", u"edge") > 0):
                                                                                            return {"prediction":3, "error":1.51877}
                                                                                        if (term_matches(title, "title", u"edge") <= 0):
                                                                                            if (item_matches(genres, "genres", u"Drama") > 0):
                                                                                                if (zipcode is None):
                                                                                                    return {"prediction": 4.83333, "error": 1.28505}
                                                                                                if (zipcode > 22103):
                                                                                                    return {"prediction":5, "error":0.57469}
                                                                                                if (zipcode <= 22103):
                                                                                                    return {"prediction":4, "error":1.28505}
                                                                                            if (item_matches(genres, "genres", u"Drama") <= 0):
                                                                                                if (timestamp > 978227687):
                                                                                                    return {"prediction":5, "error":1.56826}
                                                                                                if (timestamp <= 978227687):
                                                                                                    return {"prediction":4, "error":0.78413}
                                                                    if (timestamp <= 978215603):
                                                                        return {"prediction":3, "error":0.58872}
                                                                if (timestamp <= 978201899):
                                                                    if (term_matches(title, "title", u"lola") > 0):
                                                                        return {"prediction":4, "error":0.91271}
                                                                    if (term_matches(title, "title", u"lola") <= 0):
                                                                        if (term_matches(title, "title", u"1984") > 0):
                                                                            return {"prediction":4, "error":0.82728}
                                                                        if (term_matches(title, "title", u"1984") <= 0):
                                                                            if (term_matches(title, "title", u"terminator") > 0):
                                                                                return {"prediction":4.5, "error":5.26764}
                                                                            if (term_matches(title, "title", u"terminator") <= 0):
                                                                                return {"prediction":5, "error":0.20738}
                if (gender == "Female"):
                    if (timestamp is None):
                        return {"prediction": 4.26316, "error": 1.16276}
                    if (timestamp > 978226722):
                        if (timestamp > 978237189):
                            if (timestamp > 978238243):
                                if (term_matches(title, "title", u"1964") > 0):
                                    return {"prediction":3, "error":1.14678}
                                if (term_matches(title, "title", u"1964") <= 0):
                                    if (term_matches(title, "title", u"1996") > 0):
                                        return {"prediction":3.5, "error":5.26764}
                                    if (term_matches(title, "title", u"1996") <= 0):
                                        if (term_matches(title, "title", u"1975") > 0):
                                            return {"prediction":5, "error":0.95727}
                                        if (term_matches(title, "title", u"1975") <= 0):
                                            if (term_matches(title, "title", u"beauty") > 0):
                                                return {"prediction":5, "error":0.91271}
                                            if (term_matches(title, "title", u"beauty") <= 0):
                                                if (timestamp > 978301752):
                                                    if (timestamp > 978302153):
                                                        return {"prediction":4, "error":0.93847}
                                                    if (timestamp <= 978302153):
                                                        if (term_matches(title, "title", u"1982") > 0):
                                                            return {"prediction":4, "error":2.93426}
                                                        if (term_matches(title, "title", u"1982") <= 0):
                                                            return {"prediction":5, "error":2.07483}
                                                if (timestamp <= 978301752):
                                                    return {"prediction":4, "error":0.31268}
                            if (timestamp <= 978238243):
                                return {"prediction":5, "error":0.82845}
                        if (timestamp <= 978237189):
                            if (item_matches(genres, "genres", u"Thriller") > 0):
                                return {"prediction":4, "error":2.03402}
                            if (item_matches(genres, "genres", u"Thriller") <= 0):
                                return {"prediction":3, "error":1.17434}
                    if (timestamp <= 978226722):
                        if (term_matches(title, "title", u"1997") > 0):
                            return {"prediction":3, "error":1.35749}
                        if (term_matches(title, "title", u"1997") <= 0):
                            if (item_matches(genres, "genres", u"Adventure") > 0):
                                if (timestamp > 978153877):
                                    return {"prediction":4, "error":2.07483}
                                if (timestamp <= 978153877):
                                    return {"prediction":5, "error":2.93426}
                            if (item_matches(genres, "genres", u"Adventure") <= 0):
                                if (timestamp > 978152601):
                                    return {"prediction":5, "error":0.25805}
                                if (timestamp <= 978152601):
                                    return {"prediction":4, "error":0.77415}


def predict(gender=None,
            age_range=None,
            occupation=None,
            zipcode=None,
            title=None,
            genres=None,
            timestamp=None):
    prediction = predict_rating(gender=gender, age_range=age_range, occupation=occupation, zipcode=zipcode, title=title, genres=genres, timestamp=timestamp)
    return prediction